import os

FILE_DIR = os.path.dirname(os.path.abspath(__file__))
LOGO_PATH = os.path.join(FILE_DIR, "imgs", "tecton.png")
MELVIN_PATH = os.path.join(FILE_DIR, "imgs", "melvin.png")
STARTING_DIR = os.getcwd()
